﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Numerics;
using System.Globalization;
using System.Diagnostics;
using System.Net;
using Microsoft.Win32;
using System.Linq;

namespace GeoIPConverter
{
	class Program
	{
		// Statistics
		private static int totalLines;
		private static int processedLines;
		private static int lastReportedPercent = -1;

		// Command line options
		private static bool printProgress = true;
		private static bool verbose = false;
		private static bool waitAtEnd = true;

		private static string inPath = "in";
		private static string country4FileName;
		private static string country6FileName;
		private static string city4LocationFileName;
		private static string city4BlocksFileName;
		private static string city6FileName;
		private static string regionCodesFileName;
		private static string countriesFileName;

		private static string outPath = "out";
		private static string countryOutFileName = "geolitecountry.bin";
		private static string cityOutFileName = "geolitecity.bin";

		private static string sourceDate = "2013-10";

		static void Main(string[] args)
		{
			bool showHelp = false;
			bool download = false;
			int argi = 0;
			while (argi < args.Length)
			{
				switch (args[argi].ToLower())
				{
					case "/h":
					case "/help":
					case "/?":
						showHelp = true;
						break;
					case "/nowait":
						waitAtEnd = false;
						break;
					case "/q":
					case "/quiet":
						printProgress = false;
						break;
					case "/v":
					case "/verbose":
						verbose = true;
						break;
					case "/download":
						download = true;
						break;
					case "/in":
						inPath = args[++argi];
						break;
					case "/out":
						outPath = args[++argi];
						break;
				}
				argi++;
			}

			try
			{
				if (printProgress)
				{
					Console.WriteLine("MaxMind GeoLite Country and City database converter");
					Console.WriteLine("(c) 2012-2013 Yves Goergen - http://dev.unclassified.de/geoip");
					Console.WriteLine();

					if (verbose)
					{
						Console.WriteLine("Input path:  " + inPath);
						Console.WriteLine("Output path: " + outPath);
						Console.WriteLine();
					}
				}

				if (showHelp)
				{
					Console.WriteLine("Parameters:");
					Console.WriteLine("  /h  /help  /?    Show this help");
					Console.WriteLine("  /nowait          Don't wait when finished");
					Console.WriteLine("  /q  /quiet       Don't show progress information");
					Console.WriteLine("  /v  /verbose     Show more details and statistics");
					Console.WriteLine("  /download        Download the current archives from the MaxMind website");
					Console.WriteLine("  /in ...          Specifies the input path, default: in");
					Console.WriteLine("  /out ...         Specifies the output path, default: out");
					Console.WriteLine();
					Console.WriteLine("To just get and convert the current files, run this:");
					Console.WriteLine("  GeoIPConverter /download");
					Console.WriteLine("It will download and uncompress the archives to the subdirectory 'in' and");
					Console.WriteLine("save the converted files in the subdirectory 'out'.");
					Console.WriteLine();
					return;
				}

				// Prepare directories
				if (!Directory.Exists(inPath))
				{
					Directory.CreateDirectory(inPath);
				}
				if (!Directory.Exists(outPath))
				{
					Directory.CreateDirectory(outPath);
				}

				if (download)
				{
					if (!DownloadFiles()) return;
				}

				country4FileName = "GeoIPCountryWhois.csv";
				country6FileName = Path.GetFileName(Directory.GetFiles(inPath, "GeoIPv6*.csv").OrderByDescending(n => n).First());
				city4LocationFileName = "GeoLiteCity-Location.csv";
				city4BlocksFileName = "GeoLiteCity-Blocks.csv";
				city6FileName = "GeoLiteCityv6.csv";
				regionCodesFileName = "maxmind-region-codes.csv";
				countriesFileName = "iso-countries.csv";

				// Output file format configuration
				int headerLength = 0xa0;
				int ipv6Length = 6;   // Only store the first n bytes of an IPv6 address

				string[] allInputFiles = new string[]
				{
					countriesFileName,
					country4FileName,
					country6FileName,
					city4LocationFileName,
					city4BlocksFileName,
					city6FileName,
					regionCodesFileName
				};

				Dictionary<string, long> outFileSizes = new Dictionary<string, long>();
				Dictionary<string, int> outFileRecords = new Dictionary<string, int>();

				// Check input files exist
				DateTime latestFileTime = DateTime.MinValue;
				foreach (string inFileName in allInputFiles)
				{
					if (!File.Exists(Path.Combine(inPath, inFileName)))
					{
						ShowError("Missing input file " + inFileName);
						return;
					}

					if (inFileName != regionCodesFileName &&
						inFileName != countriesFileName)
					{
						FileInfo fi = new FileInfo(Path.Combine(inPath, inFileName));
						if (fi.LastWriteTime > latestFileTime)
						{
							latestFileTime = fi.LastWriteTime;
						}
					}
				}
				sourceDate = latestFileTime.ToString("yyyy-MM");
				Console.WriteLine("Source date: " + sourceDate);
				Console.WriteLine();

				// Count lines
				foreach (string inFileName in allInputFiles)
				{
					int weight = 1;
					if (inFileName == city4LocationFileName) weight = 4;   // Takes longer to read
					if (inFileName == countriesFileName) weight = 2;   // Will be read twice
					using (StreamReader infile = new StreamReader(Path.Combine(inPath, inFileName), Encoding.Default))
					{
						while (!infile.EndOfStream)
						{
							infile.ReadLine();
							totalLines += weight;
						}
					}
				}

				if (printProgress && verbose)
				{
					Console.WriteLine("Converting files...");
				}

				Stopwatch sw = new Stopwatch();
				sw.Start();


				// ========================================  WRITE COUNTRY DATABASE FILE  ========================================


				using (BinaryWriter outfile = new BinaryWriter(File.Create(Path.Combine(outPath, countryOutFileName)), Encoding.Default))
				{
					// Write file header
					byte[] header = Encoding.Default.GetBytes("Combined GeoLite country database v1, http://dev.unclassified.de/geoip, Data " + sourceDate + " CC-BY-SA by MaxMind, http://www.maxmind.com\r\n\r\n\r\n");
					outfile.Write(header);

					// Pad with NUL bytes to a fixed length
					while (outfile.BaseStream.Position < headerLength)
					{
						outfile.Write((byte) 0);
					}

					// Leave space for data pointers
					// (countries table starts directly after the index)
					outfile.Write((int) 0);   // country IPv4 table
					outfile.Write((int) 0);   // country IPv6 table

					// ------------------------------  Convert country names  ------------------------------

					long countryBaseAddr = outfile.BaseStream.Position;

					Dictionary<string, uint> countryOffsets = new Dictionary<string, uint>();
					using (StreamReader infile = new StreamReader(Path.Combine(inPath, countriesFileName), Encoding.Default))
					{
						ShowFileName(countriesFileName);
						int recordCount = 0;

						Regex re = new Regex(@"^(..),""(.*?)""", RegexOptions.Compiled);
						while (!infile.EndOfStream)
						{
							string line = infile.ReadLine();
							Match m = re.Match(line);
							if (m.Success)
							{
								// Remember the country table offset for each country code
								string countryCode = m.Groups[1].Value;
								// Skip unsupported country codes
								if (countryCode != "O1" &&
									countryCode != "A1" &&
									countryCode != "A2" &&
									countryCode != "AP" &&
									countryCode != "EU")
								{
									uint countryNameOffset = (uint) (outfile.BaseStream.Position - countryBaseAddr);
									if (countryNameOffset > ushort.MaxValue - 1)
									{
										ShowError("Country name offset requires more than 2 bytes, unsupported format.");
										return;
									}
									countryOffsets[countryCode] = countryNameOffset;

									// Write country code
									outfile.Write(Encoding.Default.GetBytes(countryCode));

									// Write country name
									string countryName = m.Groups[2].Value;
									if (countryName.Length > byte.MaxValue)
									{
										ShowError("Country name length requires more than 1 byte, unsupported format.");
										return;
									}
									outfile.Write((byte) countryName.Length);
									outfile.Write(Encoding.Default.GetBytes(countryName));

									recordCount++;
								}
							}

							processedLines++;
							ShowProgress();
						}

						if (recordCount == 0)
						{
							ShowError("No data found in this file, probably a wrong file or format.");
							return;
						}
						outFileRecords[countriesFileName] = recordCount;
					}

					outFileSizes[countriesFileName] = outfile.BaseStream.Position - countryBaseAddr;

					// ------------------------------  Convert country IPv4 database  ------------------------------

					// First save the base address into the header
					long addr = outfile.BaseStream.Position;
					outfile.BaseStream.Seek(headerLength, SeekOrigin.Begin);
					outfile.Write(ReverseBytes((uint) addr));
					outfile.BaseStream.Seek(addr, SeekOrigin.Begin);

					using (StreamReader infile = new StreamReader(Path.Combine(inPath, country4FileName), Encoding.Default))
					{
						ShowFileName(country4FileName);
						int recordCount = 0;
						uint lastEndIp = 0;
						int gaps = 0;

						//Regex re = new Regex(@"^\s*"".*?""\s*,\s*"".*?""\s*,\s*""(\d+)""\s*,\s*""(\d+)""\s*,\s*""(.*?)""", RegexOptions.Compiled);
						Regex re = new Regex(@"^"".*?"","".*?"",""(\d+)"",""(\d+)"",""(.*?)""", RegexOptions.Compiled);
						while (!infile.EndOfStream)
						{
							string line = infile.ReadLine();
							Match m = re.Match(line);
							if (m.Success)
							{
								string countryCode = m.Groups[3].Value;
								// Skip unsupported country codes
								if (countryCode != "O1" &&
									countryCode != "A1" &&
									countryCode != "A2" &&
									countryCode != "AP" &&
									countryCode != "EU")
								{
									uint startIp = uint.Parse(m.Groups[1].Value);
									uint endIp = uint.Parse(m.Groups[2].Value);
									if (lastEndIp != 0 && startIp != lastEndIp + 1)
									{
										gaps++;

										// Write gap filler entry
										outfile.Write(ReverseBytes(lastEndIp + 1));
										outfile.Write((byte) 0xFF);   // 2 bytes invalid country reference (0xFFFF)
										outfile.Write((byte) 0xFF);
									}
									lastEndIp = endIp;

									outfile.Write(ReverseBytes(startIp));
									//outfile.Write(ReverseBytes(endIp));

									// Write country code
									//outfile.Write(countryCode[0]);
									//outfile.Write(countryCode[1]);

									// Write country reference
									uint countryOffset;
									if (!countryOffsets.TryGetValue(countryCode, out countryOffset))
									{
										countryOffset = uint.MaxValue;
									}
									outfile.Write((byte) ((countryOffset & 0xFF00) >> 8));   // 2 bytes countryOffset (max. 64 KiB)
									outfile.Write((byte) (countryOffset & 0xFF));

									recordCount++;
								}
							}

							processedLines++;
							ShowProgress();
						}
						if (lastEndIp != 0 && lastEndIp < 0xFFFF)
						{
							// Last record doesn't go to the maximum value, add a filling record
							gaps++;

							// Write gap filler entry
							outfile.Write(ReverseBytes(lastEndIp + 1));
							outfile.Write((byte) 0xFF);   // 2 bytes invalid country reference (0xFFFF)
							outfile.Write((byte) 0xFF);
						}

						if (recordCount == 0)
						{
							ShowError("No data found in this file, probably a wrong file or format.");
							return;
						}
						outFileRecords[country4FileName] = recordCount;
					}

					outFileSizes[country4FileName] = outfile.BaseStream.Position - addr;

					// ------------------------------  Convert country IPv6 database  ------------------------------

					// First save the base address into the header
					addr = outfile.BaseStream.Position;
					outfile.BaseStream.Seek(headerLength + 4, SeekOrigin.Begin);
					outfile.Write(ReverseBytes((uint) addr));
					outfile.BaseStream.Seek(addr, SeekOrigin.Begin);

					using (StreamReader infile = new StreamReader(Path.Combine(inPath, country6FileName), Encoding.Default))
					{
						ShowFileName(country6FileName);
						int recordCount = 0;
						BigInteger lastEndIp = 0;
						int gaps = 0;

						//Regex re = new Regex(@"^\s*"".*?""\s*,\s*"".*?""\s*,\s*""(\d+)""\s*,\s*""(\d+)""\s*,\s*""(.*?)""", RegexOptions.Compiled);
						Regex re = new Regex(@"^"".*?"", "".*?"", ""(\d+)"", ""(\d+)"", ""(.*?)""", RegexOptions.Compiled);
						byte[] bytes;
						while (!infile.EndOfStream)
						{
							string line = infile.ReadLine();
							Match m = re.Match(line);
							if (m.Success)
							{
								string countryCode = m.Groups[3].Value;
								// Skip unsupported country codes
								if (countryCode != "O1" &&
									countryCode != "A1" &&
									countryCode != "A2" &&
									countryCode != "AP" &&
									countryCode != "EU")
								{
									BigInteger startIp = BigInteger.Parse(m.Groups[1].Value);
									BigInteger endIp = BigInteger.Parse(m.Groups[2].Value);
									if (lastEndIp != 0 && startIp != lastEndIp + 1)
									{
										gaps++;

										// Write gap filler entry
										bytes = (lastEndIp + 1).ToByteArray();
										for (int i = 15; i > 15 - ipv6Length; i--)
											outfile.Write(i > bytes.Length ? (byte) 0 : bytes[i]);
										outfile.Write((byte) 0xFF);   // 2 bytes invalid country reference (0xFFFF)
										outfile.Write((byte) 0xFF);
									}
									lastEndIp = endIp;

									bytes = startIp.ToByteArray();
									for (int i = 15; i > 15 - ipv6Length; i--)
										outfile.Write(i > bytes.Length ? (byte) 0 : bytes[i]);

									// Write country code
									//outfile.Write(countryCode[0]);
									//outfile.Write(countryCode[1]);

									// Write country reference
									uint countryOffset;
									if (!countryOffsets.TryGetValue(countryCode, out countryOffset))
									{
										countryOffset = uint.MaxValue;
									}
									outfile.Write((byte) ((countryOffset & 0xFF00) >> 8));   // 2 bytes countryOffset (max. 64 KiB)
									outfile.Write((byte) (countryOffset & 0xFF));

									recordCount++;
								}
							}

							processedLines++;
							ShowProgress();
						}
						if (lastEndIp != 0)
						{
							// Last record doesn't go to the maximum value, add a filling record
							gaps++;

							// Write gap filler entry
							bytes = (lastEndIp + 1).ToByteArray();
							for (int i = 15; i > 15 - ipv6Length; i--)
								outfile.Write(i > bytes.Length ? (byte) 0 : bytes[i]);
							outfile.Write((byte) 0xFF);   // 2 bytes invalid country reference (0xFFFF)
							outfile.Write((byte) 0xFF);
						}

						if (recordCount == 0)
						{
							ShowError("No data found in this file, probably a wrong file or format.");
							return;
						}
						outFileRecords[country6FileName] = recordCount;
					}

					outFileSizes[country6FileName] = outfile.BaseStream.Position - addr;
				}


				// ========================================  WRITE CITY DATABASE FILE  ========================================


				using (BinaryWriter outfile = new BinaryWriter(File.Create(Path.Combine(outPath, cityOutFileName)), Encoding.Default))
				{
					// Write file header
					byte[] header = Encoding.Default.GetBytes("Combined GeoLite city database v1, http://dev.unclassified.de/geoip, Data " + sourceDate + " CC-BY-SA by MaxMind, http://www.maxmind.com\r\n\r\n\r\n");
					outfile.Write(header);

					// Pad with NUL bytes to a fixed length
					while (outfile.BaseStream.Position < headerLength)
					{
						outfile.Write((byte) 0);
					}

					// Leave space for data pointers
					// (countries table starts directly after the index)
					outfile.Write((int) 0);   // regions table
					outfile.Write((int) 0);   // city locations table
					outfile.Write((int) 0);   // city IPv4 table
					outfile.Write((int) 0);   // city IPv6 table (currently unused, it wouldn't contain more data than the country IPv6 table)

					// ------------------------------  Convert country names  ------------------------------

					long countryBaseAddr = outfile.BaseStream.Position;

					Dictionary<string, uint> countryOffsets = new Dictionary<string, uint>();
					using (StreamReader infile = new StreamReader(Path.Combine(inPath, countriesFileName), Encoding.Default))
					{
						ShowFileName(countriesFileName);

						Regex re = new Regex(@"^(..),""(.*?)""", RegexOptions.Compiled);
						while (!infile.EndOfStream)
						{
							string line = infile.ReadLine();
							Match m = re.Match(line);
							if (m.Success)
							{
								// Remember the country table offset for each country code
								string countryCode = m.Groups[1].Value;
								// Skip unsupported country codes
								if (countryCode != "O1" &&
									countryCode != "A1" &&
									countryCode != "A2" &&
									countryCode != "AP" &&
									countryCode != "EU")
								{
									uint countryNameOffset = (uint) (outfile.BaseStream.Position - countryBaseAddr);
									if (countryNameOffset > ushort.MaxValue - 1)
									{
										ShowError("Country name offset requires more than 2 bytes, unsupported format.");
										return;
									}
									countryOffsets[countryCode] = countryNameOffset;
									// Write country code
									outfile.Write(Encoding.Default.GetBytes(countryCode));
									// Write country name
									string countryName = m.Groups[2].Value;
									if (countryName.Length > byte.MaxValue)
									{
										ShowError("Country name length requires more than 1 byte, unsupported format.");
										return;
									}
									outfile.Write((byte) countryName.Length);
									outfile.Write(Encoding.Default.GetBytes(countryName));
								}
							}

							processedLines++;
							ShowProgress();
						}
					}

					// ------------------------------  Convert regions table  ------------------------------

					// First save the base address into the header
					// Maintain 2-byte alignment for region records (required to store region pointers with half resolution)
					while ((outfile.BaseStream.Position % 2) != 0)
					{
						outfile.Write((byte) 0);
					}

					long regionBaseAddr = outfile.BaseStream.Position;
					outfile.BaseStream.Seek(headerLength, SeekOrigin.Begin);
					outfile.Write(ReverseBytes((uint) regionBaseAddr));
					outfile.BaseStream.Seek(regionBaseAddr, SeekOrigin.Begin);

					Dictionary<string, uint> regionOffsets = new Dictionary<string, uint>();
					using (StreamReader infile = new StreamReader(Path.Combine(inPath, regionCodesFileName), Encoding.Default))
					{
						ShowFileName(regionCodesFileName);
						int recordCount = 0;

						//Regex re = new Regex(@"^\s*(..)\s*,\s*""(..)""\s*,\s*""(.*?)""", RegexOptions.Compiled);
						Regex re = new Regex(@"^(..),(..),""(.*?)""", RegexOptions.Compiled);
						while (!infile.EndOfStream)
						{
							string line = infile.ReadLine();
							Match m = re.Match(line);
							if (m.Success)
							{
								// Maintain 2-byte alignment for region records (required to store region pointers with half resolution)
								while ((outfile.BaseStream.Position % 2) != 0)
								{
									outfile.Write((byte) 0);
								}

								// Remember the region table offset for each region code
								string countryCode = m.Groups[1].Value;
								string regionCode = m.Groups[2].Value;
								uint regionNameOffset = (uint) (outfile.BaseStream.Position - regionBaseAddr);
								if (regionNameOffset > ushort.MaxValue - 1)
								{
									ShowError("Region name offset requires more than 2 bytes, unsupported format.");
									return;
								}
								regionOffsets[countryCode + regionCode] = regionNameOffset;

								// Write country reference
								uint countryOffset;
								if (!countryOffsets.TryGetValue(countryCode, out countryOffset))
								{
									countryOffset = uint.MaxValue;
								}
								outfile.Write((byte) ((countryOffset & 0xFF00) >> 8));   // 2 bytes countryOffset (max. 64 KiB)
								outfile.Write((byte) (countryOffset & 0xFF));

								// Write region name
								string region = m.Groups[3].Value;
								if (region.Length > byte.MaxValue)
								{
									ShowError("Region name length requires more than 1 byte, unsupported format.");
									return;
								}
								outfile.Write((byte) region.Length);
								outfile.Write(Encoding.Default.GetBytes(region));

								recordCount++;
							}

							processedLines++;
							ShowProgress();
						}

						if (recordCount == 0)
						{
							ShowError("No data found in this file, probably a wrong file or format.");
							return;
						}
						outFileRecords[regionCodesFileName] = recordCount;
					}

					outFileSizes[regionCodesFileName] = outfile.BaseStream.Position - regionBaseAddr;

					// ------------------------------  Convert city locations table  ------------------------------

					// First save the base address into the header
					long locationBaseAddr = outfile.BaseStream.Position;
					outfile.BaseStream.Seek(headerLength + 4, SeekOrigin.Begin);
					outfile.Write(ReverseBytes((uint) locationBaseAddr));
					outfile.BaseStream.Seek(locationBaseAddr, SeekOrigin.Begin);

					Dictionary<string, uint> locationCache = new Dictionary<string, uint>();
					Dictionary<uint, uint> locationOffsets = new Dictionary<uint, uint>();
					Dictionary<string, uint> countryLocationOffsets = new Dictionary<string, uint>();
					using (StreamReader infile = new StreamReader(Path.Combine(inPath, city4LocationFileName), Encoding.Default))
					{
						ShowFileName(city4LocationFileName);
						int recordCount = 0;

						//Regex re = new Regex(@"^\s*(\d+)\s*,\s*""(.*?)""\s*,\s*""(.*?)""\s*,\s*""(.*?)""\s*,\s*"".*?""\s*,\s*(-?[0-9.]+)\s*,\s*(-?[0-9.]+)", RegexOptions.Compiled);
						Regex re = new Regex(@"^(\d+),""(.*?)"",""(.*?)"",""(.*?)"","".*?"",(-?[0-9.]+),(-?[0-9.]+)", RegexOptions.Compiled);
						while (!infile.EndOfStream)
						{
							string line = infile.ReadLine();
							Match m = re.Match(line);
							if (m.Success)
							{
								string countryCode = m.Groups[2].Value;
								string regionCode = m.Groups[3].Value;

								// Skip unsupported country codes
								if (countryCode != "O1" &&
									countryCode != "A1" &&
									countryCode != "A2" &&
									countryCode != "AP" &&
									countryCode != "EU")
								{
									// Remember the location table offset for each location ID
									uint locId = uint.Parse(m.Groups[1].Value);
									double lat = double.Parse(m.Groups[5].Value, CultureInfo.InvariantCulture);
									double lng = double.Parse(m.Groups[6].Value, CultureInfo.InvariantCulture);
									string city = m.Groups[4].Value.Trim();

									string allData = countryCode + ":" + regionCode + ":" + city + ":" + lat + ":" + lng;
									uint otherLocId;
									if (locationCache.TryGetValue(allData, out otherLocId))
									{
										// A location entry with this data already exists, reuse its offset
										locationOffsets[locId] = locationOffsets[otherLocId];
									}
									else
									{
										// This is a new location record, save it to the output file.
										// Remember its data so that it can be reused later.
										locationCache[allData] = locId;

										uint locationOffset = (uint) (outfile.BaseStream.Position - locationBaseAddr);
										if (locationOffset > 16777215)
										{
											ShowError("Location offset requires more than 3 bytes, unsupported format.");
											return;
										}
										locationOffsets[locId] = locationOffset;
										// Write location data
										uint lati = (uint) Math.Round((90 + lat) * 10000);
										uint lngi = (uint) Math.Round((180 + lng) * 10000);

										outfile.Write((byte) ((lati & 0xFF0000) >> 16));   // 3 bytes converted latitude (Y)
										outfile.Write((byte) ((lati & 0xFF00) >> 8));
										outfile.Write((byte) (lati & 0xFF));
										outfile.Write((byte) ((lngi & 0xFF0000) >> 16));   // 3 bytes converted longitude (X)
										outfile.Write((byte) ((lngi & 0xFF00) >> 8));
										outfile.Write((byte) (lngi & 0xFF));

										// Write region reference
										uint offset;
										if (regionOffsets.TryGetValue(countryCode + regionCode, out offset))
										{
											offset /= 2;   // Store region pointers with half resolution to free up one bit
										}
										else if (countryOffsets.TryGetValue(countryCode, out offset))
										{
											offset |= 0x8000;   // Set type bit to indicate that this is a country pointer
										}
										outfile.Write((byte) ((offset & 0xFF00) >> 8));   // 2 bytes region/country offset, includes type bit
										outfile.Write((byte) (offset & 0xFF));

										// Write city name
										outfile.Write((byte) city.Length);
										outfile.Write(Encoding.Default.GetBytes(city));

										if (string.IsNullOrEmpty(regionCode) && string.IsNullOrEmpty(city))
										{
											// This is a country-level location, remember it for country-code lookup
											// used in the current beta IPv6 city database
											countryLocationOffsets[countryCode] = locationOffset;
										}

										recordCount++;
									}
								}
							}

							processedLines += 4;
							ShowProgress();
						}

						if (recordCount == 0)
						{
							ShowError("No data found in this file, probably a wrong file or format.");
							return;
						}
						outFileRecords[city4LocationFileName] = recordCount;
					}

					outFileSizes[city4LocationFileName] = outfile.BaseStream.Position - locationBaseAddr;

					// ------------------------------  Convert city IPv4 database  ------------------------------

					// First save the base address into the header
					long addr = outfile.BaseStream.Position;
					outfile.BaseStream.Seek(headerLength + 8, SeekOrigin.Begin);
					outfile.Write(ReverseBytes((uint) addr));
					outfile.BaseStream.Seek(addr, SeekOrigin.Begin);

					using (StreamReader infile = new StreamReader(Path.Combine(inPath, city4BlocksFileName), Encoding.Default))
					{
						ShowFileName(city4BlocksFileName);
						int recordCount = 0;
						uint lastEndIp = 0;
						int gaps = 0;

						//Regex re = new Regex(@"^\s*""(\d+)""\s*,\s*""(\d+)""\s*,\s*""(\d+)""", RegexOptions.Compiled);
						Regex re = new Regex(@"^""(\d+)"",""(\d+)"",""(\d+)""", RegexOptions.Compiled);
						while (!infile.EndOfStream)
						{
							string line = infile.ReadLine();
							Match m = re.Match(line);
							if (m.Success)
							{
								uint locId = uint.Parse(m.Groups[3].Value);
								uint offset;
								if (locationOffsets.TryGetValue(locId, out offset))
								{
									// Missing locations must be about unsupported country codes, no further checks

									uint startIp = uint.Parse(m.Groups[1].Value);
									uint endIp = uint.Parse(m.Groups[2].Value);
									if (lastEndIp != 0 && startIp != lastEndIp + 1)
									{
										gaps++;

										// Write gap filler entry
										outfile.Write(ReverseBytes(lastEndIp + 1));
										outfile.Write((byte) 0xFF);   // 3 bytes invalid country reference (0xFFFFFF)
										outfile.Write((byte) 0xFF);
										outfile.Write((byte) 0xFF);
									}
									lastEndIp = endIp;

									outfile.Write(ReverseBytes(startIp));
									//outfile.Write(ReverseBytes(endIp));

									outfile.Write((byte) ((offset & 0xFF0000) >> 16));   // 3 bytes offset
									outfile.Write((byte) ((offset & 0xFF00) >> 8));
									outfile.Write((byte) (offset & 0xFF));

									recordCount++;
								}
							}

							processedLines++;
							ShowProgress();
						}
						if (lastEndIp != 0 && lastEndIp < 0xFFFF)
						{
							// Last record doesn't go to the maximum value, add a filling record
							gaps++;

							// Write gap filler entry
							outfile.Write(ReverseBytes(lastEndIp + 1));
							outfile.Write((byte) 0xFF);   // 3 bytes invalid country reference (0xFFFFFF)
							outfile.Write((byte) 0xFF);
							outfile.Write((byte) 0xFF);
						}

						if (recordCount == 0)
						{
							ShowError("No data found in this file, probably a wrong file or format.");
							return;
						}
						outFileRecords[city4BlocksFileName] = recordCount;
					}

					outFileSizes[city4BlocksFileName] = outfile.BaseStream.Position - addr;

					// ------------------------------  Convert city IPv6 database  ------------------------------

					// First save the base address into the header
					addr = outfile.BaseStream.Position;
					outfile.BaseStream.Seek(headerLength + 12, SeekOrigin.Begin);
					outfile.Write(ReverseBytes((uint) addr));
					outfile.BaseStream.Seek(addr, SeekOrigin.Begin);

					using (StreamReader infile = new StreamReader(Path.Combine(inPath, city6FileName), Encoding.Default))
					{
						ShowFileName(city6FileName);
						int recordCount = 0;
						BigInteger lastEndIp = 0;
						int gaps = 0;

						Regex re = new Regex(@"^.*?,.*?,(\d+),(\d+),(.*?),.*?,.*?,(-?[0-9.]+),(-?[0-9.]+)", RegexOptions.Compiled);
						byte[] bytes;
						while (!infile.EndOfStream)
						{
							string line = infile.ReadLine();
							Match m = re.Match(line);
							if (m.Success)
							{
								string countryCode = m.Groups[3].Value;
								// Skip unsupported country codes
								if (countryCode != "O1" &&
									countryCode != "A1" &&
									countryCode != "A2" &&
									countryCode != "AP" &&
									countryCode != "EU")
								{
									BigInteger startIp = BigInteger.Parse(m.Groups[1].Value);
									BigInteger endIp = BigInteger.Parse(m.Groups[2].Value);
									if (lastEndIp != 0 && startIp != lastEndIp + 1)
									{
										gaps++;

										// Write gap filler entry
										bytes = (lastEndIp + 1).ToByteArray();
										for (int i = 15; i > 15 - ipv6Length; i--)
											outfile.Write(i > bytes.Length ? (byte) 0 : bytes[i]);
										outfile.Write((byte) 0xFF);   // 3 bytes invalid city reference (0xFFFFFF)
										outfile.Write((byte) 0xFF);
										outfile.Write((byte) 0xFF);
									}
									lastEndIp = endIp;

									// Only store the first 64 bits of the IPv6 address, the interface part is not assigned
									bytes = startIp.ToByteArray();
									for (int i = 15; i > 15 - ipv6Length; i--)
										outfile.Write(i > bytes.Length ? (byte) 0 : bytes[i]);

									// Write country reference
									uint offset;
									if (!countryLocationOffsets.TryGetValue(countryCode, out offset))
									{
										ShowError("Country location for IPv6 record of " + countryCode + " is unknown.");
										return;
									}
									outfile.Write((byte) ((offset & 0xFF0000) >> 16));   // 3 bytes offset
									outfile.Write((byte) ((offset & 0xFF00) >> 8));
									outfile.Write((byte) (offset & 0xFF));

									recordCount++;
								}
							}

							processedLines++;
							ShowProgress();
						}
						if (lastEndIp != 0)
						{
							// Last record doesn't go to the maximum value, add a filling record
							gaps++;

							// Write gap filler entry
							bytes = (lastEndIp + 1).ToByteArray();
							for (int i = 15; i > 15 - ipv6Length; i--)
								outfile.Write(i > bytes.Length ? (byte) 0 : bytes[i]);
							outfile.Write((byte) 0xFF);   // 3 bytes invalid city reference (0xFFFFFF)
							outfile.Write((byte) 0xFF);
							outfile.Write((byte) 0xFF);
						}

						if (recordCount == 0)
						{
							ShowError("No data found in this file, probably a wrong file or format.");
							return;
						}
						outFileRecords[city6FileName] = recordCount;
					}

					outFileSizes[city6FileName] = outfile.BaseStream.Position - addr;
				}

				// Clear progress line from the console
				processedLines = totalLines + 1;
				ShowProgress();

				if (printProgress)
				{
					Console.WriteLine();
					Console.WriteLine("Conversion successfully finished in " +
						Math.Round((double) sw.ElapsedMilliseconds / 1000, 1).ToString(CultureInfo.InvariantCulture) +
						" seconds.");

					if (verbose)
					{
						Console.WriteLine();
						Console.WriteLine("Data size of each input file in the resulting database file:");
						foreach (string inFileName in allInputFiles)
						{
							Console.Write(inFileName.PadRight(26));
							Console.Write(outFileSizes[inFileName].ToString("N0", CultureInfo.InvariantCulture).PadLeft(10));
							Console.Write(" of ");
							Console.Write(new FileInfo(Path.Combine(inPath, inFileName)).Length
								.ToString("###,###,##0", CultureInfo.InvariantCulture).PadLeft(10));
							Console.Write(" bytes");
							Console.Write(outFileRecords[inFileName].ToString("N0", CultureInfo.InvariantCulture).PadLeft(11));
							Console.WriteLine(" records");
						}
					}
				}
			}
			finally
			{
				if (printProgress && waitAtEnd)
				{
					//System.Threading.Thread.Sleep(1000);
					Console.WriteLine();
					Console.WriteLine("Press any key to exit...");
					Console.ReadKey();
				}
			}
		}

		private static byte[] StringBytes(string str, int length)
		{
			byte[] bytes = new byte[length];
			int i;
			for (i = 0; i < str.Length; i++)
			{
				bytes[i] = (byte) str[i];
			}
			while (i < length)
			{
				bytes[i] = 0x20;   // Space
				i++;
			}
			return bytes;
		}

		private static uint ReverseBytes(uint i)
		{
			return
				(i & 0xFF000000) >> 24 |
				(i & 0x00FF0000) >> 8 |
				(i & 0x0000FF00) << 8 |
				(i & 0x000000FF) << 24;
		}

		private static void ShowFileName(string fileName)
		{
			if (printProgress)
			{
				if (Console.CursorTop > 0 && lastReportedPercent >= 0)
					Console.CursorTop--;
				Console.WriteLine(fileName + "                      ");
				lastReportedPercent = -1;
			}
		}

		private static void ShowProgress()
		{
			if (printProgress)
			{
				int percent = processedLines * 100 / totalLines;
				if (processedLines > totalLines)
				{
					percent = 101;
				}
				if (percent > lastReportedPercent)
				{
					if (Console.CursorTop > 0 && lastReportedPercent >= 0)
						Console.CursorTop--;
					if (percent <= 100)
					{
						Console.WriteLine(percent + "% completed...");
					}
					else
					{
						Console.WriteLine("                    ");
						Console.CursorTop--;
					}
					lastReportedPercent = percent;
				}
			}
		}

		private static bool DownloadFiles()
		{
			if (printProgress)
			{
				Console.WriteLine("Downloading files... (expecting ~30 MB, this may take a while)");
			}

			DateTime today = DateTime.Today;
			DateTime firstTuesday = new DateTime(today.Year, today.Month, 1);
			while (firstTuesday.DayOfWeek != DayOfWeek.Tuesday)
			{
				firstTuesday = firstTuesday.AddDays(1);
			}

			string baseUrl = "http://geolite.maxmind.com/download/geoip/database/";
			if (!DownloadFile(baseUrl + "GeoIPCountryCSV.zip", "GeoIPCountryCSV.zip")) return false;
			if (!DownloadFile(baseUrl + "GeoIPv6.csv.gz", "GeoIPv6.csv.gz")) return false;
			if (!DownloadFile(baseUrl + "GeoLiteCity_CSV/GeoLiteCity_" + firstTuesday.ToString("yyyyMMdd") + ".zip", "GeoLiteCity.zip")) return false;
			if (!DownloadFile(baseUrl + "GeoLiteCityv6-beta/GeoLiteCityv6.csv.gz", "GeoLiteCityv6.csv.gz")) return false;
			if (!DownloadFile("http://dev.maxmind.com/static/maxmind-region-codes.csv", "maxmind-region-codes.csv")) return false;
			if (!DownloadFile("http://dev.unclassified.de/files/source/geoip/iso-countries.csv", "iso-countries.csv")) return false;

			if (printProgress)
			{
				Console.WriteLine("Decompressing files...");
			}

			// Try to find 7-Zip
			string sevenZipPath = null;
			try
			{
				sevenZipPath = Registry.CurrentUser.OpenSubKey("Software\\7-Zip").GetValue("Path") as string;
				sevenZipPath = Path.Combine(sevenZipPath, "7z.exe");
				if (!File.Exists(sevenZipPath))
					sevenZipPath = null;
			}
			catch
			{
			}
			if (sevenZipPath == null)
			{
				// Not installed, try portable binary in the search path
				sevenZipPath = "7za.exe";
			}

			if (!DecompressFile(sevenZipPath, "GeoIPCountryCSV.zip")) return false;
			if (!DecompressFile(sevenZipPath, "GeoIPv6.csv.gz")) return false;
			if (!DecompressFile(sevenZipPath, "GeoLiteCity.zip")) return false;
			if (!DecompressFile(sevenZipPath, "GeoLiteCityv6.csv.gz")) return false;

			return true;
		}

		private static bool DownloadFile(string url, string fileName)
		{
			HttpWebRequest request = (HttpWebRequest) WebRequest.Create(url);
			request.UserAgent = "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:16.0) Gecko/20100101 Firefox/16.0";

			try
			{
				using (HttpWebResponse response = (HttpWebResponse) request.GetResponse())
				{
					if (response.StatusCode != HttpStatusCode.OK)
					{
						ShowError("File " + fileName + " could not be downloaded, HTTP code " + (int) response.StatusCode);
						return false;
					}

					using (Stream stream = response.GetResponseStream())
					using (FileStream fs = new FileStream(Path.Combine(inPath, fileName), FileMode.Create, FileAccess.ReadWrite, FileShare.None))
					{
						stream.CopyTo(fs);
					}
				}
			}
			catch (WebException ex)
			{
				HttpWebResponse response = (HttpWebResponse) ex.Response;
				ShowError("File " + fileName + " could not be downloaded, HTTP code " + (int) response.StatusCode);
				return false;
			}
			return true;
		}

		private static bool DecompressFile(string sevenZipPath, string archiveFileName)
		{
			try
			{
				ProcessStartInfo psi = new ProcessStartInfo(sevenZipPath, "e -y " + archiveFileName);
				psi.WorkingDirectory = inPath;
				psi.UseShellExecute = false;
				if (!printProgress || !verbose)
				{
					psi.RedirectStandardOutput = true;
				}
				Process process = Process.Start(psi);
				process.WaitForExit();
				if (printProgress && verbose)
				{
					Console.WriteLine();
				}
				if (process.ExitCode != 0)
				{
					ShowError("File " + archiveFileName + " could not be decompressed, 7-Zip return code " + process.ExitCode);
					return false;
				}

				return true;
			}
			catch (FileNotFoundException)
			{
				ShowError("7-Zip not found, downloaded files cannot be decompressed");
				return false;
			}
		}

		private static void ShowError(string text)
		{
			ConsoleColor fg = Console.ForegroundColor;
			Console.ForegroundColor = ConsoleColor.Red;
			Console.Error.WriteLine("Error: " + text);
			Console.ForegroundColor = fg;
		}
	}
}
